use sdl2::event::Event;
use sdl2::keyboard::Keycode;
use sdl2::pixels::Color;
use sdl2::rect::Rect;
use sdl2::render::Texture;
use sdl2::render::TextureCreator;
use sdl2::render::TextureQuery;
use sdl2::ttf;
use sdl2::ttf::Font;
use sdl2::video::WindowContext;

fn string_err(e: impl ToString) -> String {
    e.to_string()
}

const WIDTH: u32 = 800;
const HEIGHT: u32 = 600;

fn main() -> Result<(), String> {
    let context = sdl2::init()?;
    let video = context.video()?;
    let ttf_context = ttf::init().map_err(string_err)?;

    let window = video
        .window("medit", WIDTH, HEIGHT)
        .position_centered()
        .allow_highdpi()
        .build()
        .map_err(string_err)?;

    let mut canvas = window.into_canvas().build().map_err(string_err)?;
    let texture_creator = canvas.texture_creator();

    let font = ttf_context.load_font("resources/fonts/ttf/FiraCode-VF.ttf", 128)?;

    canvas.set_draw_color(Color::WHITE);
    canvas.clear();

    let mut event_pump = context.event_pump()?;
    let mut x = 0;
    'running: loop {
        for event in event_pump.poll_iter() {
            match event {
                Event::Quit { .. }
                | Event::KeyDown {
                    keycode: Some(Keycode::Escape),
                    ..
                } => break 'running,
                Event::KeyDown {
                    keycode: Some(k), ..
                } if is_drawable_key(k) => {
                    canvas.clear();
                    let texture = string_texture(&font, &texture_creator, k.name())?;
                    let target = target_rect(x, 0, &texture);
                    x += target.width() as i32;
                    canvas.copy(&texture, None, Some(target))?;
                }
                _ => {}
            }
        }
        canvas.present();
        std::thread::sleep(std::time::Duration::new(0, 1_000_000_000u32 / 60));
    }
    Ok(())
}

fn is_drawable_key(key: Keycode) -> bool {
    let key = key as i32;
    (key >= Keycode::A as i32 && key <= Keycode::Z as i32)
        || (key >= Keycode::Num0 as i32 && key <= Keycode::Num9 as i32)
}

fn string_texture<'f, 't>(
    font: &Font<'f, 'f>,
    texture_creator: &'t TextureCreator<WindowContext>,
    string: impl AsRef<str>,
) -> Result<Texture<'t>, String> {
    let surface = font
        .render(string.as_ref())
        .blended(Color::BLACK)
        .map_err(string_err)?;
    texture_creator
        .create_texture_from_surface(surface)
        .map_err(string_err)
}

fn target_rect(x: i32, y: i32, texture: &Texture) -> Rect {
    let TextureQuery { width, height, .. } = texture.query();

    Rect::new(x, y, width, height)
}
